# -*- coding: utf-8 -*-

# @Time    : 2020/9/8 23:51
# @Email   : 986798607@qq.com
# @Software: PyCharm
# @License: BSD 3-Clause
# -*- coding: utf-8 -*-

import numpy as np
import sklearn
import sklearn.utils
from mgetool.tool import parallelize
from sklearn.utils import check_array

from multiego.base_ego import BaseEgo


def search_space(*arg):
    """
    Generate grid.

    Note
    ------
        Parameters should be no more than 6

    Parameters
    ----------
    arg: list of np.ndarray
        Examples:
            arg = [
            np.arange(0.1,0.35,0.1),
            np.arange(0.1, 2.1, 0.5),
            np.arange(0,1.3,0.3),
            np.array([0.5,1,1.2,1.3]),]

    Returns
    -------
    result: np.ndarray
    """
    meshes = np.meshgrid(*arg)
    meshes = [_.ravel() for _ in meshes]
    meshes = np.array(meshes).T
    return meshes


class Ego(BaseEgo):
    """
    EGO (Efficient global optimization).

    References
    -----------
        Jones, D. R., Schonlau, M. & Welch, W. J. Efficient global optimization of expensive black-box functions. J.
        Global Optim. 13, 455–492 (1998)

    Examples
    ----------

        >>> searchspace_list = [
        ...    np.arange(0.1,0.35,0.1),
        ...    np.arange(0.1, 1.3, 0.3),
        ...    np.arange(0.1, 2.1, 0.5),
        ...    np.arange(0,1.3,0.3),
        ...    np.arange(0,7.5,1.5),
        ...    np.arange(0,7.5,1.5),
        ...    np.arange(800, 1300, 50),
        ...    np.arange(200, 600, 40),
        ...    np.array([20, 80, 138, 250]),]

        >>> searchspace = search_space(*searchspace_list)
        >>> me = Ego(searchspace, X, y, 500, SVR(), n_jobs=8)
        >>> result = me.rank()

    Attributes
    ------------
    mean_std: np.ndarray
        mean and std
    predict_y: np.ndarray
        predict_y

    Notes
    ------
        Result is 2 dimentions array
        1st column = sequence number,
        2nd part = your searchspace,
        3rd part = 'mean,std,ego,kg,maxp,sequentially'

    """

    def __init__(self, regclf=None, searchspace=None, X=None, y=None, number=500, n_jobs=2):
        """

        Parameters
        ----------
        searchspace: np.ndarray of shape (n_sample_pre, n_feature)
            searchspace with the same ``n_feature`` with X,
            Custom or generate by .search_space() function.
        X: np.ndarray of shape (n_sample_train, n_feature)
            X data (2D).
        y: np.ndarray of shape (n_sample_train, 1)
            y data (1D).
        number: int>100
            Repeat number, default is 500.
        regclf: sklearn.estimator
            sklearn module, with "fit" and "predict".
        n_jobs: int
            Parallelize number.
        """
        super(Ego, self).__init__()

        self.n_jobs = n_jobs
        self.searchspace = searchspace
        self.X = X
        self.y = y
        self.regclf = regclf
        self.meanandstd_all = []
        self.predict_y_all = []
        self.number = number
        self.mean_std = None
        self.rank=self.egosearch

    def fit(self, searchspace=None, X=None, y=None, *args):
        """

        Parameters
        ----------
        searchspace: np.ndarray of shape (n_sample_pre, n_feature)
            searchspace with the same ``n_feature`` with X,
            Custom or generate by .search_space() function.
        X: np.ndarray of shape (n_sample_train, n_feature)
            X data (2D).
        y: np.ndarray of shape (n_sample_train, 1)
            y data (1D).

        """
        assert hasattr(self.regclf, "fit")
        assert hasattr(self.regclf, "predict")

        self.searchspace = self.searchspace if searchspace is None else searchspace
        self.X = self.X if X is None else X
        self.y = self.y if y is None else y
        searchspace = self.searchspace
        X = self.X
        y = self.y

        njobs = self.n_jobs
        regclf0 = self.regclf
        assert searchspace is not None and X is not None and y is not None, "searchspace, X, y should be np.array"
        check_array(X, ensure_2d=True, force_all_finite=True)
        check_array(y, ensure_2d=False, force_all_finite=True)
        check_array(searchspace, ensure_2d=True, force_all_finite=True)
        assert X.shape[1] == searchspace.shape[1]

        def fit_parllize(random_state):
            data_train, y_train = sklearn.utils.resample(X, y, n_samples=None, replace=True,
                                                         random_state=random_state)
            regclf0.fit(data_train, y_train)
            predict_data = regclf0.predict(searchspace)
            predict_data.ravel()
            return predict_data

        predict_y = parallelize(n_jobs=njobs, func=fit_parllize, iterable=range(self.number))
        predict_y = np.array(predict_y).T

        self.predict_y = predict_y

        self.meanandstd()

    def meanandstd(self, predict_y=None):
        """calculate meanandstd"""
        if predict_y is not None:
            self.predict_y = predict_y
        if not hasattr(self, "predict_y"):
            raise NotImplemented("Please fit first")
        if self.predict_y is None:
            raise NotImplemented("Please fit first")

        mean_std = super().meanandstd(self.predict_y)

        self.mean_std = mean_std
        return self.mean_std

    def egosearch(self, searchspace=None, mean_std=None, rankway="ego", return_type="pd", flexibility=0, y=None,
                  fraction=1000):
        """
        Result is 2 dimentions array
        1st column = sequence number,2nd part = your searchspace,3rd part = mean,std,ego,kg,maxp,sequentially.

        Parameters
        ----------
        y: np.ndarray of shape (n_sample_true, 1)
            y data (1D).
        searchspace : np.ndarray of shape (n_sample_pre, n_feature)
            search space
            ["ego","kg","maxp","No"]
        return_type:str
            numpy.ndarray or pandas.DataFrame
        mean_std: np.ndarray of shape (n_sample_pre, n_feature)
            mean_std of n times of prediction on search space.
            First column is mean and second is std.
        rankway : str
            ["ego","kg","maxp","No"]
        fraction:int
            choice top n_sample/fraction.
        flexibility:float
            Flexibility to calculate EI, the bigger flexibility, the more search space Ei >0.
        """

        y = self.y if y is None else y
        searchspace = self.searchspace if searchspace is None else searchspace
        mean_std = self.mean_std if mean_std is None else mean_std

        if mean_std is None:
            self.fit()
            mean_std = self.mean_std

        return super().egosearch(y=y, searchspace=searchspace, mean_std=mean_std, rankway=rankway,
                                 return_type=return_type, flexibility=flexibility, fraction=fraction)


if __name__ == "__main__":
    from sklearn.datasets import load_boston
    import numpy as np
    from sklearn.svm import SVR

    #####model1#####
    model = SVR()
    ###

    #####model2#####
    # parameters = {'C': [0.1, 1, 10]}
    # model = GridSearchCV(SVR(), parameters)
    ###

    X, y = load_boston(return_X_y=True)
    X = X[:, :5]  # (简化计算，示意)
    searchspace_list = [
        np.arange(0.01, 1, 0.1),
        np.array([0, 20, 30, 50, 70, 90]),
        np.arange(1, 10, 1),
        np.array([0, 1]),
        np.arange(0.4, 0.6, 0.02),
    ]
    searchspace = search_space(*searchspace_list)

    me = Ego(model, searchspace, X, y, 100, n_jobs=6)
    me.fit()
    re = me.egosearch()



import unittest

import numpy as np
import pandas as pd
from sklearn.datasets import load_boston
from sklearn.model_selection import GridSearchCV
from sklearn.svm import SVR

from multiego.base_ego import BaseEgo
from multiego.ego import search_space


class MyTestCase(unittest.TestCase):
    def setUp(self) -> None:
        parameters = {'C': [0.1, 1, 10]}
        model = GridSearchCV(SVR(), parameters)
        ###

        X, y = load_boston(return_X_y=True)
        X = X[:, :5]  # (简化计算，示意)
        searchspace_list = [
            np.arange(0.01, 1, 0.2),
            np.array([0, 20, 30]),
            np.arange(1, 10, 2),
            np.array([0, 1]),
            np.arange(0.4, 0.6, 0.1),
        ]

        searchspace = search_space(*searchspace_list)
        self.y = y
        self.searchspace = searchspace

    def test_something(self):
        mean_std = np.random.random((self.searchspace.shape[0], 1000))
        be = BaseEgo()

        try:
            rank = be.egosearch(self.y, self.searchspace, mean_std, rankway="ego", return_type="pd", flexibility=20)

        except AssertionError:
            pass
        try:
            rank = be.egosearch(self.y, self.searchspace, mean_std, rankway="ego", return_type="pd", flexibility=100)
        except UserWarning:
            pass


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python
# coding:utf-8
"""
This is one general method to calculate Efficient global optimization with multiple target ,
There are no restrictions on the type of X and model.

"""
import warnings
from itertools import chain, combinations

import numpy as np
import pandas as pd
from mgetool.tool import batch_parallelize

warnings.filterwarnings("ignore")


def search_space(*arg):
    """

    Parameters
    ----------
    arg: list of np.ndarray
        Examples:
            arg = [
            np.arange(0.1,0.35,0.1),
            np.arange(0.1, 2.1, 0.5),
            np.arange(0,1.3,0.3),
            np.array([0.5,1,1.2,1.3]),]

    Returns
    -------
    np.ndarray

    """
    meshes = np.meshgrid(*arg)
    meshes = [_.ravel() for _ in meshes]
    meshes = np.array(meshes).T
    return meshes


class BaseMultiplyEgo:
    """
    EGO (Efficient global optimization).
    References:
        Jones, D. R., Schonlau, M. & Welch, W. J. Efficient global optimization of expensive black-box functions. J.
        Global Optim. 13, 455–492 (1998)

    Attributes
    ------------
    front_point: np.ndarray
        front_point.
    front_point_index: np.ndarray
        front_point index in x.
    Ei: np.ndarray
        Ei
    Pi: np.ndarray
        Pi
    L: np.ndarray
        Pi
    """

    def __init__(self, n_jobs=2, up=True):
        """

        Parameters
        ----------
        n_jobs:int
            parallelize number.

        up:bool
            The pareto front face is going to be up or down for pareto front point.

        """

        self.n_jobs = n_jobs
        self.rank = self.egosearch
        self.up = up

    @staticmethod
    def meanandstd(predict_y):
        mean = np.mean(predict_y, axis=1)
        std = np.std(predict_y, axis=1)
        data_predict = np.column_stack((mean, std))
        # print(data_predict.shape)
        return data_predict

    def get_meanandstd_all(self, predict_ys):
        meanandstd = []
        for i in range(predict_ys.shape[-1]):
            meanandstd_i = self.meanandstd(predict_ys[:, :, i])
            meanandstd.append(meanandstd_i)
        return meanandstd

    def pareto_front_point(self, y, sign=None):
        m = y.shape[0]
        n = y.shape[1]
        if sign is None:
            sign = np.array([1] * n)
        y *= sign
        front_point_index = []
        for i in range(m):
            data_new = y[i, :].reshape(1, -1) - y
            data_max = np.max(data_new, axis=1)
            data_in = np.min(data_max)
            if data_in >= 0:
                front_point_index.append(i)
        front_point = y[front_point_index, :].T
        self.front_point = front_point
        self.front_point_index = front_point_index
        return front_point

    def CalculateL(self, meanandstd_all):
        front_y = self.front_point
        meanstd = np.array(meanandstd_all)
        meanstd = meanstd[:, :, 0].T
        alll = []
        for front_y_i in front_y.T:
            l_i = meanstd - front_y_i
            alll.append(l_i)
        dmin = np.array(alll)

        dmin2 = np.min(np.abs(dmin), axis=0)

        dmin3 = np.min(dmin2, axis=1)

        # dmin3 = np.sqrt(np.sum(dmin2**2,axis=1)) # another method.

        dmin3[np.where(dmin3 < 0)[0]] = 0

        self.L = dmin3
        return dmin3

    def CalculateEi(self, y, meanandstd_all=None, predict_y_all=None, sign=None, flexibility=None):
        """EI value"""
        if flexibility is None:
            flexibility = np.array(np.zeros((y.shape[1], 1)))
        else:
            warnings.warn(
                "``Flexibility`` means reduction of y boundary, Please use it if you know what you are doing.")
            flexibility = np.array(flexibility).reshape(-1, 1)

        self.pareto_front_point(y, sign)
        self.CalculatePi(predict_y_all, flexibility=flexibility)
        self.CalculateL(meanandstd_all)
        Ei = self.L * self.Pi
        self.Ei = Ei
        return Ei

    def scrap(self, fp):
        """This function is add the up step to paretofront face.
        For the front, the steps can go down or up, we add up to get tighter requirements.
        """
        # max problem
        if fp.shape[0] == 1:
            raise NotImplemented("For single problem, we dont implement up-pareto front face methods.")
        if fp.shape[0] == 2:
            fp_index = np.argsort(fp[0, :])
            fp = fp[:, fp_index]
            new_fp = np.vstack((fp[0, :][1:], fp[1, :][:-1]))

        else:
            # nps= []
            # for coi in combinations(range(fp.shape[1]),fp.shape[0]):
            #     new_point = fp[:,coi]
            #     nps.append(np.max(new_point,axis=1))
            # nps=np.array(nps)
            # new_fp = self.pareto_front_point(nps)
            raise NotImplemented("For multi problem more than 2, we dont implemente up-pareto front face methods. please set ``up = False``")

        return new_fp

    def CalculatePi(self, predict_y_all, flexibility):
        """PI value"""

        njobs = self.n_jobs

        if self.up:
            front_y = self.scrap(self.front_point)
        else:
            front_y = self.front_point

        front_y -= flexibility

        def tile_func(i):
            tile = 0
            for front_y_i in front_y.T:
                big = i - front_y_i
                big_bool = np.max(big, axis=1) < 0
                tile |= big_bool
            return tile

        tile_all = batch_parallelize(n_jobs=njobs, func=tile_func, iterable=predict_y_all, batch_size=100)
        pi = np.sum(1 - np.array(tile_all), axis=1) / predict_y_all.shape[1]

        self.Pi = pi

        return pi

    def egosearch(self, y, predict_y_all, meanandstd_all=None, searchspace=None, return_type="pd", flexibility=None,
                  fraction=1000, sign=None):
        """
        Result is 2 dimensions array.
        1st column = sequence number,\n
        2nd part = your search space,\n
        3rd part = Pi,L,Ei,sequentially.

        Parameters
        ----------
        y: np.ndarray of shape (n_sample_train, n_model)
            true train y.
        searchspace : np.ndarray of shape (n_sample_pre, n_feature)
            search space, for base_multiego, searchspace is not used and just as one placeholder.
        fraction: int
            choice top n_sample/fraction
        return_type:str
            numpy.ndarray or pandas.DataFrame
        meanandstd_all: list of np.ndarray
            Not required force.
            n_model meanandstd, Each meanandstd is np.ndarray of shape (n_sample_pre,2)
        predict_y_all: np.ndarray of shape (n_sample_pre,n_times,n_model)
            ys.
        sign:np.ndarray of shape (n_model,)
            Each element must be -1 or 1.
            sign to define the max problem or min problem.

        flexibility: List[float]
            Flexibility to calculate PI, the bigger flexibility, the more search space Pi >0. for each y.

        Returns
        ----------
        table:np.ndarray (2d), pd.Dateframe

        """
        if flexibility is None:
            flexibility = np.array(np.zeros((y.shape[1], 1)))
        else:
            warnings.warn(
                "``Flexibility`` means reduction of y boundary, Please use it if you know what you are doing.")
            flexibility = np.array(flexibility).reshape(-1, 1)

        bianhao = np.arange(0, predict_y_all.shape[0]).reshape(-1, 1)

        if searchspace is None:
            searchspace = np.arange(0, predict_y_all.shape[0]).reshape(-1, 1)

        if meanandstd_all is None:
            meanandstd_all = self.get_meanandstd_all(predict_y_all)

        self.CalculateEi(y, meanandstd_all, predict_y_all, sign=sign, flexibility=flexibility)

        assert not np.all(self.Ei <= 1e-10), "All the Ei (and Pi) score is 0, This is invalid calculation. " \
                                             "Please try these methods:\n" \
                                             "1. Improve your model precision, especially near the expected scope for y. " \
                                             "For example, for max proplem, the point near the maximum y should be accurate by model.\n" \
                                             "2. Make sure your search space near the training space.\n" \
                                             "3. If the above methods are still unable to solve, add flexibility to find point by reduction of y boundary. ()\n"

        result1 = np.column_stack((bianhao, searchspace, *meanandstd_all, self.Pi, self.L, self.Ei))

        max_paixu = np.argsort(-result1[:, -1])

        if max_paixu.size >= fraction:
            select_number = max_paixu[:int(max_paixu.size / fraction)]
        else:
            print("grid is smaller than fraction, the ``fraction`` is ignored.")
            select_number = max_paixu

        result1 = result1[select_number]

        if return_type == "pd":
            result1 = pd.DataFrame(result1)
            fea = ["feature%d" % i for i in range(searchspace.shape[1])]
            meanstds = ["meanstd%d" % i for i in range(sum([i.shape[1] for i in meanandstd_all]))]
            name = ["number"] + fea + meanstds + ["Pi", "L", "Ei"]
            result1.columns = name
        self.result = result1
        return self.result



# -*- coding: utf-8 -*-

# @Time    : 2020/9/8 23:51
# @Email   : 986798607@qq.com
# @Software: PyCharm
# @License: BSD 3-Clause
# -*- coding: utf-8 -*-

"""This is one general method to calculate Efficient global optimization,
This one is with no restrictions on the type of X and model.

"""
import warnings

import numpy as np
import pandas as pd
from scipy import stats


class BaseEgo:
    """
    EGO (Efficient global optimization).

    References:
        Jones, D. R., Schonlau, M. & Welch, W. J.
        Efficient global optimization of expensive black-box functions. J.
        Global Optim. 13, 455–492 (1998)

    Examples:

        >>>me = BaseEgo()
        >>>result = me.rank(y=y, mean_std=mean_std)

    """

    def __init__(self):
        self.rank = self.egosearch

    @staticmethod
    def meanandstd(predict_y):
        """calculate meanandstd."""
        mean = np.mean(predict_y, axis=1)
        std = np.std(predict_y, axis=1)
        data_predict = np.column_stack((mean, std))
        print(data_predict.shape)
        return data_predict

    @staticmethod
    def CalculateEi(y, mean_std0, flexibility=0.0):

        """calculate EI."""
        ego = (mean_std0[:, 0] - (max(y) - flexibility)) / (mean_std0[:, 1])
        ei_ego = mean_std0[:, 1] * ego * stats.norm.cdf(ego) + mean_std0[:, 1] * stats.norm.pdf(ego)
        kg = (mean_std0[:, 0] - max(max(mean_std0[:, 0]), max(y) - flexibility)) / (mean_std0[:, 1])
        ei_kg = mean_std0[:, 1] * kg * stats.norm.cdf(kg) + mean_std0[:, 1] * stats.norm.pdf(kg)
        max_P = stats.norm.cdf(ego)
        ei = np.column_stack((mean_std0, ei_ego, ei_kg, max_P))
        print('Ego is done.')
        return ei

    def egosearch(self, y, mean_std, searchspace=None, rankway="ego", return_type="pd", flexibility=0, fraction=1000, ):
        """
        Result is 2 dimensions array.
        1st column = sequence number,\n
        2nd part = your search space,\n
        3rd part = mean,std,ego,kg,maxp,sequentially.

        Parameters
        ----------
        y: np.ndarray of shape (n_sample_train, 1)
            train y.
        mean_std: np.ndarray of shape (n_sample_pre, n_feature)
            mean_std of n times of prediction on search space.
            First column is mean and second is std.
        rankway : str
            ["ego","kg","maxp","No"]
            resort the result by rankway name.
        searchspace : np.ndarray of shape (n_sample_pre, n_feature)
            search space, the search space in BaseEgo is not used, just as one placeholder for corresponding .
        return_type: str
            "pd" or "np"
        fraction:int
            choice top n_sample/fraction.
        flexibility:float
            Flexibility to calculate EI, the bigger flexibility, the more search space Ei >0.
            make sure less than max(y) for max problem.

        Returns
        ----------
        table:np.ndarray (2d), pd.Dateframe

        """
        if flexibility !=0:
            warnings.warn("``Flexibility`` means reduction of y boundary, Please use it if you know what you are doing.")

        reverse = False
        if rankway not in ['ego', 'kg', 'maxp', 'no', 'No']:
            print('Don\'t kidding me,checking rankway=what?\a')
        else:

            result = self.CalculateEi(y, mean_std, flexibility=flexibility)
            bianhao = np.arange(0, len(result)).reshape(-1,1)
            if searchspace is None:
                searchspace=bianhao
            result1 = np.column_stack((bianhao, searchspace, result))

            # for all score: [-3,-2,-1] could check by one [-3]
            assert not np.all(result1[:, -3] <= 1e-10), "All the ego score is 0, This is invalid calculation. " \
                                                   "Please try these methods:\n" \
                                                   "1. Improve your model precision, especially near the expected scope for y. " \
                                                   "For example, for max proplem, the point near the maximum y should be accurate by model.\n" \
                                                   "2. Make sure your search space near the training space.\n" \
                                                   "3. If the above methods are still unable to solve, add flexibility to find point by reduction of y boundary. ()\n"

            if rankway == "ego":
                if reverse:
                    paixu = np.argsort(result1[:, -3])
                else:
                    paixu = np.argsort(-result1[:, -3])

            elif rankway == "kg":
                if reverse:
                    paixu = np.argsort(result1[:, -2])
                else:
                    paixu = np.argsort(-result1[:, -2])

            elif rankway == "maxp":
                if reverse:
                    paixu = np.argsort(result1[:, -1])
                else:
                    paixu = np.argsort(-result1[:, -1])

            else:
                paixu = bianhao

            if paixu.size >= fraction:
                select_number = paixu[:int(paixu.size / fraction)]
            else:
                print(
                    "search space grid number {} is smaller than fraction, the ``fraction`` parameter is ignored.".format(
                        paixu.size))
                select_number = paixu

            result1 = result1[select_number]

            if return_type != "pd":
                return result1
            else:
                result1 = pd.DataFrame(result1)
                fea = ["feature%d" % i for i in range(searchspace.shape[1])]
                mean_stds = ["mean_std%d" % i for i in range(mean_std.shape[1])]
                name = ["number"] + fea + mean_stds + ['ego', 'kg', 'maxp']
                result1.columns = name
                return result1



import unittest

import numpy as np
# @Time    : 2020/11/25 14:50
# @Email   : 986798607@qq.com
# @Software: PyCharm
# @License: BSD 3-Clause
from sklearn.datasets import load_boston
from sklearn.linear_model import Lasso
from sklearn.model_selection import GridSearchCV
from sklearn.svm import SVR

from multiego.base_multiplyego import search_space, BaseMultiplyEgo


# -*- coding: utf-8 -*-


class MyTestCase(unittest.TestCase):
    def setUp(self) -> None:
        parameters = {'C': [0.1, 1, 10]}

        model = GridSearchCV(SVR(), parameters)
        parameters = {'alpha': [0.1, 1, 10]}
        model2 = GridSearchCV(Lasso(), parameters)

        X, y = load_boston(return_X_y=True)
        X = X[:, :5]  # (简化计算，示意)
        searchspace_list = [
            np.arange(0.01, 1, 0.2),
            np.array([0, 20, 30]),
            np.arange(1, 10, 2),
            np.array([0, 1]),
            np.arange(0.4, 0.6, 0.1),
        ]

        searchspace = search_space(*searchspace_list)
        self.model1 = model
        self.model2 = model2

        self.y = np.concatenate([y.reshape(-1, 1), 2 * y.reshape(-1, 1)], axis=1)
        self.searchspace = searchspace

    def test_something(self):
        me = BaseMultiplyEgo(n_jobs=1)
        mean_std = [np.random.random((self.searchspace.shape[0], 2)), np.random.random((self.searchspace.shape[0], 2))]
        predict_y_all = np.random.random((self.searchspace.shape[0], 1000, 2))
        rank = me.egosearch(self.y, predict_y_all,meanandstd_all=mean_std, flexibility=80)


if __name__ == '__main__':
    unittest.main()



import gc
import warnings
from collections.abc import Iterable

import numpy as np
import sklearn.utils
from mgetool.tool import parallelize
from sklearn.utils import check_array

from multiego.base_multiplyego import BaseMultiplyEgo

warnings.filterwarnings("ignore")


def search_space(*arg):
    """

    Parameters
    ----------
    arg: list of np.ndarray
        Examples:
            arg = [
            np.arange(0.1,0.35,0.1),
            np.arange(0.1, 2.1, 0.5),
            np.arange(0,1.3,0.3),
            np.array([0.5,1,1.2,1.3]),]

    Returns
    -------
    np.ndarray

    """
    meshes = np.meshgrid(*arg)
    meshes = [_.ravel() for _ in meshes]
    meshes = np.array(meshes).T
    return meshes


class MultiplyEgo(BaseMultiplyEgo):
    """
    Multi-target EGO (Efficient global optimization).

    References
    -----------
        Jones, D. R., Schonlau, M. & Welch, W. J. Efficient global optimization of expensive black-box functions. J.
        Global Optim. 13, 455–492 (1998)
    Examples
    -----------

        >>> searchspace_list = [
        ...    np.arange(0.1,0.35,0.1),
        ...    np.arange(0.1, 1.3, 0.3),
        ...    np.arange(0.1, 2.1, 0.5),
        ...    np.arange(0,1.3,0.3),
        ...    np.arange(0,7.5,1.5),
        ...    np.arange(0,7.5,1.5),
        ...    np.arange(800, 1300, 50),
        ...    np.arange(200, 600, 40),
        ...    np.array([20, 80, 138, 250]),]

        >>> searchspace = search_space(*searchspace_list)

        >>> me = MutilplyEgo([svr, gpr], searchspace, X, y, 500, n_jobs=8)

        >>> me.fit()

        >>> select_number = me.rank()
        >>> result = me.result

    Attributes
    ------------
    mean_std_all: np.ndarray
        mean and std.
    predict_y_all: np.ndarray
        predict_y.
    front_point: np.ndarray
        front_point.
    front_point_index: np.ndarray
        front_point index in x.
    Ei: np.ndarray
        Ei
    Pi: np.ndarray
        Pi
    L: np.ndarray
        Pi
    """

    def __init__(self, regclf, searchspace, X, y, number=1000, feature_slice=None, n_jobs=2, up=True):
        """

        Parameters
        ----------
        searchspace:np.ndarray of shape (n_sample_pre,n_feature)
            custom or generate by .search_space() function.
        X: np.ndarray of shape (n_sample_train, n_feature)
            X data (2D).
        y: np.ndarray of shape (n_sample_train, n_model)
            y data (2D).
        number:int>100
            repeat number,default is 1000.
        regclf: List of callable
            sklearn methods, with "fit" and "predict".
            The same number with the number of target y.
        feature_slice: None,list of tuple.
            The same number with the number of target y,
            Aand each tuple contains the features index for each regclf.
        n_jobs:int
            parallelize number.
        """
        super(MultiplyEgo, self).__init__(n_jobs=n_jobs,up=up)
        self.rank = self.egosearch

        check_array(X, ensure_2d=True, force_all_finite=True)
        check_array(y, ensure_2d=False, force_all_finite=True)
        check_array(searchspace, ensure_2d=True, force_all_finite=True)
        assert X.shape[1] == searchspace.shape[1]
        self.searchspace = searchspace
        self.X = X
        self.y = y
        self.sign = []

        assert isinstance(regclf, Iterable)
        assert len(list(regclf)) >= 2
        self.regclf = list(regclf)
        self.dim = len(list(regclf))

        if feature_slice is None:
            feature_slice = tuple([tuple(range(X.shape[1]))] * self.dim)
        assert isinstance(feature_slice, (tuple, list))
        assert isinstance(feature_slice[1], (tuple, list))
        assert self.dim == len(feature_slice) == self.y.shape[1]
        self.feature_slice = feature_slice

        self.Ei = np.zeros_like(searchspace[:, 1])
        self.Pi = np.zeros_like(searchspace[:, 1])
        self.L = np.zeros_like(searchspace[:, 1])
        self.front_point = np.zeros_like(self.y[:, 1])
        self.number = number
        self.center = np.zeros_like(searchspace[:, 1])
        self.rank = self.egosearch


    def _fit(self, x, y, searchspace0, regclf0):
        def fit_parllize(random_state):
            data_train, y_train = sklearn.utils.resample(x, y, n_samples=None, replace=True,
                                                         random_state=random_state)
            regclf0.fit(data_train, y_train)
            predict_data = regclf0.predict(searchspace0)
            predict_data.ravel()
            return predict_data

        predict_dataj = parallelize(n_jobs=self.n_jobs, func=fit_parllize, iterable=range(self.number))

        return np.array(predict_dataj).T

    def fit(self, regclf_number=None):
        """

        Parameters
        ----------
        regclf_number: int,list

            the running sklearn methods, default is all of regclf.

        Returns
        -------
        meanandstd: np.ndarray
             mean and std predicted y value with number(1000).
        """

        if regclf_number is None:
            contain = list(range(self.dim))
        elif isinstance(regclf_number, int):
            contain = [regclf_number]
        elif isinstance(regclf_number, (list, tuple)):
            contain = regclf_number
        else:
            raise TypeError()
        meanandstd = []
        predict_y_all = []
        for i, feature_slicei, yi, regclfi in zip(range(self.dim), self.feature_slice, self.y.T, self.regclf):
            if i in contain:
                predict_y = np.array(self._fit(self.X[:, feature_slicei], yi, self.searchspace[:, feature_slicei],
                                               regclfi))
                predict_y_all.append(predict_y)

                meanandstd_i = self.meanandstd(predict_y)

                del predict_y
                gc.collect()

                meanandstd.append(meanandstd_i)
            else:
                pass

        self.mean_std_all = meanandstd
        self.predict_y_all = np.array(predict_y_all).transpose((1,2,0))
        return self.mean_std_all

    def egosearch(self, searchspace=None, fraction=1000, return_type="pd", predict_y_all=None,
                  meanandstd_all=None, flexibility=None,
                  y=None, sign=None):
        """
        Result is 2 dimensions array.
        1st column = sequence number,\n
        2nd part = your search space,\n
        3rd part = Pi,L,Ei,sequentially.

        Parameters
        ----------
        y: np.ndarray of shape (n_sample_train, n_model)
            true train y.
        searchspace : np.ndarray of shape (n_sample_pre, n_feature)
            search space
        fraction: int
            choice top n_sample/fraction
        return_type:str
            numpy.ndarray or pandas.DataFrame
        meanandstd_all: list of np.ndarray
            Not required force.
            n_model meanandstd, Each meanandstd is np.ndarray of shape (n_sample_pre,2)
        predict_y_all: np.ndarray of shape (n_sample_pre,n_times,n_model)
            ys.
        sign:np.ndarray of shape (n_model,)
            Each element must be -1 or 1.
            sign to define the max problem or min problem.

        flexibility: List[float]
            Flexibility to calculate PI, the bigger flexibility, the more search space Pi >0. for each y.

        Returns
        ----------
        table:np.ndarray (2d), pd.Dateframe

        """
        y = self.y if y is None else y
        if searchspace is not None:
            self.searchspace = searchspace
        if predict_y_all is not None:
            self.predict_y_all = predict_y_all
        if meanandstd_all is not None:
            self.mean_std_all = meanandstd_all

        if hasattr(self, 'mean_std_all') and hasattr(self, 'predict_y_all'):
            pass
        elif hasattr(self, 'predict_y_all'):
            self.mean_std_all = self.get_meanandstd_all(self.predict_y_all)
        else:
            self.fit()

        return super().egosearch(y, self.predict_y_all, searchspace= self.searchspace,meanandstd_all= self.mean_std_all,
                                 return_type="pd", fraction=fraction, flexibility=flexibility, sign=sign)

# #
# #
# #
# #
# #
# #
# #
# #
# #
# #
# #
# #
# #
# #
# #
# #
# #
# #
# #
# #
# #
# # #!/usr/bin/python
# # # coding:utf-8
# #
# #
# # """
# # sample
# # """
# # import gc
# # import warnings
# # from collections.abc import Iterable
# #
# # import numpy as np
# # import pandas as pd
# # import sklearn.utils
# # from mgetool.tool import parallelize
# # from sklearn.utils import check_array
# #
# # warnings.filterwarnings("ignore")
# #
# #
# # def search_space(*arg):
# #     """
# #
# #     Parameters
# #     ----------
# #     arg: list of np.ndarray
# #         Examples:
# #             arg = [
# #             np.arange(0.1,0.35,0.1),
# #             np.arange(0.1, 2.1, 0.5),
# #             np.arange(0,1.3,0.3),
# #             np.array([0.5,1,1.2,1.3]),]
# #
# #     Returns
# #     -------
# #     np.ndarray
# #
# #     """
# #     meshes = np.meshgrid(*arg)
# #     meshes = [_.ravel() for _ in meshes]
# #     meshes = np.array(meshes).T
# #     return meshes
# #
# #
# # class MultiplyEgo:
# #     """
# #     EGO (Efficient global optimization).
# #     References:
# #         Jones, D. R., Schonlau, M. & Welch, W. J. Efficient global optimization of expensive black-box functions. J.
# #         Global Optim. 13, 455–492 (1998)
# #     Examples:
# #
# #         searchspace_list = [
# #             np.arange(0.1,0.35,0.1),
# #             np.arange(0.1, 1.3, 0.3),
# #             np.arange(0.1, 2.1, 0.5),
# #             np.arange(0,1.3,0.3),
# #             np.arange(0,7.5,1.5),
# #             np.arange(0,7.5,1.5),
# #             np.arange(800, 1300, 50),
# #             np.arange(200, 600, 40),
# #             np.array([20, 80, 138, 250]),]
# #
# #         searchspace = search_space(*searchspace_list)
# #
# #         me = MutilplyEgo([svr, gpr], searchspace, X, y, 500, n_jobs=8)
# #
# #         me.fit()
# #
# #         select_number = me.Rank()
# #
# #         result = self.result
# #
# #     """
# #
# #     def __init__(self, regclf, searchspace, X, y, number=1000, feature_slice=None, n_jobs=2):
# #         """
# #
# #         Parameters
# #         ----------
# #         searchspace:np.ndarray
# #             custom or generate by .search_space() function.
# #         X: np.ndarray
# #             X data (2D).
# #         y: np.ndarray
# #             y data (2D).
# #         number:int>100
# #             repeat number,default is 1000.
# #         regclf: list of callable
# #             sklearn methods, with "fit" and "predict".
# #             The same number with the number of target y.
# #         feature_slice: None,list of tuple.
# #             The same number with the number of target y,
# #             and the tuple contains the features for regclf.
# #         n_jobs:int
# #             parallelize number.
# #         """
# #
# #         self.n_jobs = n_jobs
# #         check_array(X, ensure_2d=True, force_all_finite=True)
# #         check_array(y, ensure_2d=False, force_all_finite=True)
# #         check_array(searchspace, ensure_2d=True, force_all_finite=True)
# #         assert X.shape[1] == searchspace.shape[1]
# #         self.searchspace = searchspace
# #         self.X = X
# #         self.y = y
# #         self.sign = []
# #
# #         assert isinstance(regclf, Iterable)
# #         assert len(list(regclf)) >= 2
# #         self.regclf = list(regclf)
# #         self.dim = len(list(regclf))
# #
# #         if feature_slice is None:
# #             feature_slice = tuple([tuple(range(X.shape[1]))] * self.dim)
# #         assert isinstance(feature_slice, (tuple, list))
# #         assert isinstance(feature_slice[1], (tuple, list))
# #         assert self.dim == len(feature_slice) == self.y.shape[1]
# #         self.feature_slice = feature_slice
# #
# #         self.Ei = np.zeros_like(searchspace[:, 1])
# #         self.Pi = np.zeros_like(searchspace[:, 1])
# #         self.L = np.zeros_like(searchspace[:, 1])
# #         self.front_point = np.zeros_like(self.y[:, 1])
# #         self.number = number
# #         self.center = np.zeros_like(searchspace[:, 1])
# #
# #     def _fit(self, x, y, searchspace0, regclf0):
# #         def fit_parllize(random_state):
# #             data_train, y_train = sklearn.utils.resample(x, y, n_samples=None, replace=True,
# #                                                          random_state=random_state)
# #             regclf0.fit(data_train, y_train)
# #             predict_data = regclf0.predict(searchspace0)
# #             predict_data.ravel()
# #             return predict_data
# #
# #         njobs = self.n_jobs
# #         number = self.number
# #
# #         predict_y = parallelize(n_jobs=njobs, func=fit_parllize, iterable=range(number))
# #
# #         return np.array(predict_y)
# #
# #     @staticmethod
# #     def _mean_and_std(predict_y):
# #         mean = np.mean(predict_y, axis=0)
# #         std = np.std(predict_y, axis=0)
# #         data_predict = np.column_stack((mean, std))
# #         # print(data_predict.shape)
# #         return data_predict
# #
# #     def fit(self, regclf_number=None):
# #         """
# #
# #         Parameters
# #         ----------
# #         regclf_number: int,list
# #
# #             the running sklearn methods, default is all of regclf.
# #
# #         Returns
# #         -------
# #         meanandstd: np.ndarray
# #              mean and std predicted y value with number(1000).
# #         """
# #
# #         if regclf_number is None:
# #             contain = list(range(self.dim))
# #         elif isinstance(regclf_number, int):
# #             contain = [regclf_number]
# #         elif isinstance(regclf_number, (list, tuple)):
# #             contain = regclf_number
# #         else:
# #             raise TypeError()
# #         meanandstd = []
# #         predict_y_all = []
# #         for i, feature_slicei, yi, regclfi in zip(range(self.dim), self.feature_slice, self.y.T, self.regclf):
# #             if i in contain:
# #                 predict_y = np.array(self._fit(self.X[:, feature_slicei], yi, self.searchspace[:, feature_slicei],
# #                                                regclfi))
# #                 predict_y_all.append(predict_y)
# #
# #                 meanandstd_i = self._mean_and_std(predict_y)
# #
# #                 del predict_y
# #                 gc.collect()
# #
# #                 meanandstd.append(meanandstd_i)
# #             else:
# #                 pass
# #
# #         self.meanandstd_all = meanandstd
# #         self.predict_y_all = np.array(predict_y_all).T
# #         return self.meanandstd_all
# #
# #     def pareto_front_point(self):
# #         sign = self.sign
# #         y = self.y
# #         m = y.shape[0]
# #         n = y.shape[1]
# #         if not sign:
# #             sign = np.array([1] * n)
# #         y *= sign
# #         front_point = []
# #         for i in range(m):
# #             data_new = y[i, :].reshape(1, -1) - y
# #             data_max = np.max(data_new, axis=1)
# #             data_in = np.min(data_max)
# #             if data_in >= 0:
# #                 front_point.append(i)
# #         self.front_point_index = front_point
# #         self.front_point = self.y[front_point, :].T
# #         return front_point
# #
# #     # def CalculateEI2(spath):
# #     #     predict_y_all = spath.predict_y_all
# #     #     front_y = spath.pareto_front_point()
# #     #     front_y = spath.y[front_y, :].T
# #     #     meanstd = spath.meanandstd_all
# #     #     pi_all = 1
# #     #     center = []
# #     #     for y_i, meanstd_i in zip(front_y, meanstd):
# #     #         std_ = meanstd_i[:, 1]
# #     #         mean_ = meanstd_i[:, 0]
# #     #         y_max = max(y_i)
# #     #         upper_bound = (mean_ - y_max) / std_
# #     #         numerator = [integrate.quad(lambda x: 1 / np.sqrt(2 * np.pi) * x * np.exp(-0.5 * x ** 2),
# #     #                                     a=upper_bound_i, b=np.inf,
# #     #                                     full_output=0)[0] for upper_bound_i in upper_bound]
# #     #         pi_i_denominator = stats.norm.cdf(upper_bound)
# #     #         center_i = numerator / pi_i_denominator
# #     #         pi_all *= pi_i_denominator
# #     #         center.append(center_i)
# #     #     center = np.array(center)
# #     #     spath.Pi = pi_all
# #     #     spath.center = center
# #     #     L = np.array([[spatial.distance.euclidean(y_j, j) for j in center.T] for y_j in front_y.T])
# #     #     L_min = np.min(L, axis=0)
# #     #     spath.L = L_min
# #
# #     def CalculateL(self, meanandstd_all=None):
# #         front_y = self.pareto_front_point()
# #         front_y = self.y[front_y, :].T
# #         if meanandstd_all is not None:
# #             self.meanandstd_all = meanandstd_all
# #         meanstd = np.array(self.meanandstd_all)
# #         meanstd = meanstd[:, :, 0].T
# #         alll = []
# #         for front_y_i in front_y.T:
# #             l_i = meanstd - front_y_i
# #             alll.append(l_i)
# #         dmin = np.array(alll)
# #
# #         dmin2 = np.min(np.abs(dmin), axis=0)
# #
# #         dmin3 = np.min(dmin2, axis=1)
# #
# #         #        dmin3 = np.sqrt(np.sum(dmin2**2,axis=1))
# #
# #         dmin3[np.where(dmin3 < 0)[0]] = 0
# #
# #         self.L = dmin3
# #         return dmin3
# #
# #     def CalculateEi(self, meanandstd_all=None, predict_y_all=None):
# #         """EI value"""
# #         self.CalculatePi(predict_y_all)
# #         self.CalculateL(meanandstd_all)
# #         Ei = self.L * self.Pi
# #         self.Ei = Ei
# #         return Ei
# #
# #     def CalculatePi(self, predict_y_all=None):
# #         """PI value"""
# #         if predict_y_all is not None:
# #             self.predict_y_all = predict_y_all
# #
# #         njobs = self.n_jobs
# #         front_y = self.pareto_front_point()
# #         front_y = self.y[front_y, :].T
# #
# #         def tile_func(i, front_y0):
# #             tile = 0
# #             for front_y_i in front_y0.T:
# #                 big = i - front_y_i
# #                 big_bool = np.max(big, axis=1) < 0
# #                 tile |= big_bool
# #             return tile
# #
# #         tile_all = parallelize(n_jobs=njobs, func=tile_func, iterable=self.predict_y_all, front_y0=front_y)
# #         pi = np.sum(1 - np.array(tile_all), axis=1) / self.number
# #         self.Pi = pi
# #
# #         return pi
# #
# #     def Rank(self, fraction=1000, return_type="pd", predict_y_all=None, meanandstd_all=None):
# #         """
# #
# #         Parameters
# #         ----------
# #         fraction: int
# #             choice top n_sample/fraction
# #         return_type:str
# #
# #             numpy.ndarray or pandas.DataFrame
# #
# #         meanandstd_all:np.ndarray, None
# #
# #         predict_y_all：np.ndarray
# #
# #         Returns
# #         -------
# #         rank
# #         """
# #         if predict_y_all is not None:
# #             self.predict_y_all = predict_y_all
# #         if meanandstd_all is not None:
# #             self.meanandstd_all = meanandstd_all
# #         bianhao = np.arange(0, self.searchspace.shape[0])
# #         if hasattr(self, 'meanandstd_all') and hasattr(self, 'predict_y_all'):
# #             self.CalculateEi(self.meanandstd_all, self.predict_y_all)
# #         else:
# #             self.fit()
# #             self.CalculateEi(self.meanandstd_all, self.predict_y_all)
# #
# #         del self.predict_y_all
# #         gc.collect()
# #
# #         result1 = np.column_stack((bianhao, self.searchspace, *self.meanandstd_all, self.Pi, self.L, self.Ei))
# #         max_paixu = np.argsort(-result1[:, -1])
# #         select_number = max_paixu[:int(max_paixu.size / fraction)]
# #         result1 = result1[select_number]
# #
# #         if return_type == "pd":
# #             result1 = pd.DataFrame(result1)
# #             fea = ["feature%d" % i for i in range(self.searchspace.shape[1])]
# #             meanstds = ["meanstd%d" % i for i in range(sum([i.shape[1] for i in self.meanandstd_all]))]
# #             name = ["number"] + fea + meanstds + ["Pi", "L", "Ei"]
# #             result1.columns = name
# #         self.result = result1
# #         return self.result
# #
